/******************************************************************************\

                Copyright (c) 2013 Chris Burns & Warren Hunt
                            All rights reserved.

       Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
   
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   * Neither the name of the <organization> nor the
     names of its contributors may be used to endorse or promote products
     derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
    OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
              EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

      This code accompanies the Journal of Computer Graphics Techniques 
  (http://jcgt.org) article titled "The Visibility Buffer: A Cache-Friendly 
    Approach to Deferred Shading". These files were automatically generated 
                    by the system described in the article.

\******************************************************************************/

#define __LANCELOT__
#define SUPER_SAMPLED_AA           100
#define TRIANGLE_SAMPLED_AA        200
#define STENCIL_AA                 440
#define WIREFRAME_AA               450
#define AA_MODE SUPER_SAMPLED_AA
#define MIPMAP_LOD_BIAS      0.5
#define SAMPLES_PER_PIXEL    4
#define INV_SPP              0.25
#define TILE_SIZE            128
#define TILE_SIZE_X          16
#define TILE_SIZE_Y          8
#define INSTANCE_ID_MASK     1023
#define INSTANCE_ID_BITS     10
#define MAX_SHADING_SAMPLES_PER_TILE 512

cbuffer __SystemConstants
{
   matrix RasterToWorld;
   float4 CameraPositionWorldSpace;
   uint2 ImageSize;
   uint NumLights;
   uint ShaderID;
};

float CalculateTextureLOD(float filterWidth, Texture2D tex)
{
   uint N_LEVELS, Width, Height;
   tex.GetDimensions(0, Width, Height, N_LEVELS);
   return (N_LEVELS - 1.0f + MIPMAP_LOD_BIAS) + log2(filterWidth);
}

cbuffer __SurfaceConstants {
   float4 diffuseColor;
   float4 specularColor;
   float specularExp;
};
SamplerState LinearSampler
{
   Filter = MIN_MAG_MIP_LINEAR;
   AddressU = Wrap;
   AddressV = Wrap;
};

StructuredBuffer<uint> __lightList;
#define L_TYPE_MASK  0x0000000f
#define L_INDEX_MASK (~L_TYPE_MASK)
struct __shadePointLight_Constants {
   uint2 lightIntensityAndRadius;
   uint2 worldSpacePosition;
};
StructuredBuffer<__shadePointLight_Constants> shadePointLight_ConstantsArray;
StructuredBuffer<matrix> shadePointLight_TransformsArray;
/// GROUP SHARED DATA for AA_MODE == PER_TRIANGLE
/// Overallocate, allows omission of range checking code
typedef uint4 ShadingRecord;
groupshared ShadingRecord shadingRecords[MAX_SHADING_SAMPLES_PER_TILE];
groupshared uint nShadingSamples;
groupshared uint __perPixelPixels[TILE_SIZE];
groupshared uint __nPerPixelPixels;
groupshared uint __perSamplePixels[TILE_SIZE];
groupshared uint __nPerSamplePixels;
/// GROUP SHARED DATA for AA_MODE == STENCIL_AA
// List of pixels that require per-sample shading We encode
// two 16-bit x/y coordinates in one uint to save shared memory space
groupshared uint sPerSamplePixels[TILE_SIZE];
groupshared uint sNumPerSamplePixels;
uint __packPixelCoords(uint2 coords)
{ return coords.y << 16 | coords.x; }
uint2 __unpackPixelCoords(uint coords)
{ return uint2(coords & 0xFFFF, coords >> 16); }
// RGBA 16-bit per component packed into a uint2 per texel
float4 __unpackRGBA16(uint2 e)
{ return float4(f16tof32(e), f16tof32(e >> 16)); }
uint2 __packRGBA16(float4 c)
{ return f32tof16(c.rg) | (f32tof16(c.ba) << 16); }

struct VertexInput_Storage
{
   float4 position;
   uint2 tangent_texCoord_u;
   uint2 bitangent_texCoord_v;
};

struct VertexInput
{
   float4 position  ;
   float4 tangent   ;
   float4 bitangent ;
   float4 texCoord  ;
};

struct VertexOutput
{
   float4 position  ;
   float4 tangent   ;
   float4 bitangent ;
   float4 texCoord  ;

   float4 clipSpacePosition : SV_POSITION;
};

struct Surface {
   float4 worldSpacePosition;
   float4 normal;
   float4 diffuse;
   float4 specular;
};
struct Radiance {
   float4 flux;
   float4 direction;
};

VertexOutput shadeVertex(matrix ObjectToWorld, VertexInput vIn)
{
   VertexOutput vOut;
   vOut.position  = vIn.position;
   vOut.texCoord  = vIn.texCoord;
   vOut.tangent   = mul( ObjectToWorld, float4(vIn.tangent.xyz, 0.0)  );
   vOut.bitangent = mul( ObjectToWorld, float4(vIn.bitangent.xyz, 0.0));
   return vOut;
}

Surface shadeSurface(float FilterWidth, int2 PixelCoord, float2 ImageCoord, matrix ObjectToWorld, VertexOutput v)
{
   Surface s;
   s.diffuse.xyz = diffuseColor;  //< Uniform value
   s.specular.xyz = specularColor;  //< Uniform value
   s.normal.xyz = normalize(cross(v.tangent.xyz, v.bitangent.xyz));
   s.worldSpacePosition.xyz = mul(ObjectToWorld, float4(v.position.xyz, 1.0f)).xyz;
   s.specular.w = specularExp;
   return s;
}

float3 shadeBRDF(Surface s, Radiance r)
{
   float3 L = r.direction.xyz;          //< Normalized, world-space, away from surface
   float3 I = r.flux.xyz;
   float3 N = s.normal.xyz;             //< World-space
   float3 P = s.worldSpacePosition.xyz; //< World-space

   float3 E = normalize(CameraPositionWorldSpace.xyz - P);
   float3 H = normalize(L+E);

   float3 kD = s.diffuse.xyz * I;
   float3 kS = s.specular.xyz * I * pow(max(dot(N, H), 0.0f), s.specular.w);

   return kD + kS;
}
/// __color is multisampled, use __writeSample(uint2, uint, float4)
///  to compute the sample index, downconvert if necessary and write
RWStructuredBuffer<uint2> __color;

void __writePixel(uint PixelIndex, float4 color)
{ __color[PixelIndex] = __packRGBA16(color); }
void __blendPixel(uint PixelIndex, float4 color)
{
   float4 dst = __unpackRGBA16(__color[PixelIndex]);
   __color[PixelIndex] = __packRGBA16(color + dst);
}

Texture2DMS<uint>                     __visibilityBuffer;
StructuredBuffer<VertexInput_Storage> __vertexUberBuffer;
Buffer<uint4>                           __indexUberBuffer;
StructuredBuffer<matrix>              __transformUberBuffer;
Texture2D<uint>                       __msaaStencil;
Buffer<uint4>                         __instanceToIndexMap;

Buffer<int> __offsets;
Buffer<int> __tiles;

Buffer<float4> __vertexUberBuffer_Position;
Buffer<float4> __vertexUberBuffer_Tangent;
Buffer<float4> __vertexUberBuffer_Bitangent;

VertexInput __loadVertex(uint uberBufferIndex)
{
   VertexInput vIn;
   vIn.position  = __vertexUberBuffer_Position   [uberBufferIndex];
   vIn.tangent   = __vertexUberBuffer_Tangent   [uberBufferIndex];
   vIn.bitangent = __vertexUberBuffer_Bitangent   [uberBufferIndex];

   vIn.texCoord.x = vIn.tangent.w;
   vIn.texCoord.y = vIn.bitangent.w;
   vIn.texCoord.zw = float2(1.0f, 1.0f);

   return vIn;
}


Radiance shadePointLight(uint2 lightIntensityAndRadius, uint2 worldSpacePosition, matrix LightToWorld, Surface s)
{
   float3 P = float3 //LightToWorld._m03_m13_m23;
      ( f16tof32(worldSpacePosition.x)
      , f16tof32(worldSpacePosition.x >> 16)
      , f16tof32(worldSpacePosition.y));

   float3 I = float3
      ( f16tof32(lightIntensityAndRadius.x)
      , f16tof32(lightIntensityAndRadius.x >> 16)
      , f16tof32(lightIntensityAndRadius.y));

   float radius = f16tof32(lightIntensityAndRadius.y >> 16);

   float3 V = P - s.worldSpacePosition.xyz;
   float  dSquared = dot(V, V);
   float3 L = normalize(V);
   float falloff = max(0.0f, 1.0f - dSquared / (radius * radius));

   Radiance r;
   r.direction.xyz = L;
   r.flux.xyz = falloff * I * max(dot(s.normal.xyz, L), 0.0f);
   return r;
}

float __dot3(float3 a, float3 b)
{ return mad(a.x, b.x, mad(a.y, b.y, a.z*b.z)); }
float3 __cross3(float3 a, float3 b)
{ return mad(a, b.yzx, -a.yzx * b).yzx;}

float3 __intersect(float3 p0, float3 p1, float3 p2, float3 o, float3 d)
{
   float3 eo =  o - p0;
   float3 e2 = p2 - p0;
   float3 e1 = p1 - p0;
   float3 r  = __cross3(d, e2);
   float3 s  = __cross3(eo, e1);
   float iV  = 1.0f / __dot3(r, e1);
   float V1  = __dot3(r, eo);
   float V2  = __dot3(s,  d);
   float b   = V1 * iV;
   float c   = V2 * iV;
   float a   = 1.0f - b - c;
   return float3(a, b, c);
}

VertexInput __loadAndInterpolateVertex(uint iOffset, uint vOffset, uint tID, matrix ObjectToWorld, float2 pixelCoord, out float filterWidth)
{
   uint4 vIndices = __indexUberBuffer[iOffset + tID];
   uint baseVertex = vOffset;

   VertexInput v0 = __loadVertex( baseVertex + vIndices.x );
   VertexInput v1 = __loadVertex( baseVertex + vIndices.y );
   VertexInput v2 = __loadVertex( baseVertex + vIndices.z );

   float4 p0 = mul(ObjectToWorld, v0.position);
   float4 p1 = mul(ObjectToWorld, v1.position);
   float4 p2 = mul(ObjectToWorld, v2.position);

   float3 d
      = RasterToWorld._m00_m10_m20  * (pixelCoord.x + 0.5f)
      + RasterToWorld._m01_m11_m21  * ((ImageSize.y-pixelCoord.y-1) + 0.5f)
      + RasterToWorld._m03_m13_m23 ;

   /// Compute the barycentric coordinates
   float3 H  = __intersect(p0.xyz, p1.xyz, p2.xyz, RasterToWorld._m02_m12_m22 , d);

   VertexInput vIn;
   vIn.position  = mad(v0.position,  H.x, mad(v1.position,  H.y, (v2.position  * H.z)));
   vIn.tangent   = mad(v0.tangent,   H.x, mad(v1.tangent,   H.y, (v2.tangent   * H.z)));
   vIn.bitangent = mad(v0.bitangent, H.x, mad(v1.bitangent, H.y, (v2.bitangent * H.z)));
   vIn.texCoord  = mad(v0.texCoord,  H.x, mad(v1.texCoord,  H.y, (v2.texCoord  * H.z)));

   /// Compute texture LOD values based on the texCoord attribute
   /// If the LOD is not used by the user shader(s), then this all gets dead-code eliminated
   float3 dx
      = RasterToWorld._m00_m10_m20  * (pixelCoord.x + 1.5f)
      + RasterToWorld._m01_m11_m21  * ((ImageSize.y-pixelCoord.y-1) + 0.5f)
      + RasterToWorld._m03_m13_m23 ;
   float3 dy
      = RasterToWorld._m00_m10_m20  * (pixelCoord.x + 0.5f)
      + RasterToWorld._m01_m11_m21  * ((ImageSize.y-pixelCoord.y-1) + 1.5f)
      + RasterToWorld._m03_m13_m23 ;
   float3 Hx = __intersect(p0.xyz, p1.xyz, p2.xyz, RasterToWorld._m02_m12_m22 , dx);
   float3 Hy = __intersect(p0.xyz, p1.xyz, p2.xyz, RasterToWorld._m02_m12_m22 , dy);
   float2 tCoordDX = mad(v0.texCoord.xy, Hx.x, mad(v1.texCoord.xy, Hx.y, (v2.texCoord.xy * Hx.z)));
   float2 tCoordDY = mad(v0.texCoord.xy, Hy.x, mad(v1.texCoord.xy, Hy.y, (v2.texCoord.xy * Hy.z)));
   float dudx = vIn.texCoord.x - tCoordDX.x, dvdx = vIn.texCoord.y - tCoordDX.y;
   float dudy = vIn.texCoord.x - tCoordDY.x, dvdy = vIn.texCoord.y - tCoordDY.y;
   filterWidth = 0.5f * (sqrt(dudx*dudx + dvdx*dvdx) + sqrt(dudy*dudy + dvdy*dvdy));

   return vIn;
}

float4 __shadeDifferentialSurface
   ( VertexInput vIn
   , matrix ObjectToWorld
   , int2 PixelCoord
   , float2 ImageCoord
   , float FilterWidth)
{
   VertexOutput v = shadeVertex(ObjectToWorld, vIn);
   Surface surface = shadeSurface(FilterWidth, PixelCoord, ImageCoord, ObjectToWorld, v);

   float4 color = float4(0.0f, 0.0f, 0.0f, 1.0f);
   uint lightIndex = 0;
   while (lightIndex < NumLights)
   {
      [unroll] for (uint n = 0; n < 8; ++n, ++lightIndex) 
      {
         uint lightID = __lightList[lightIndex];
         uint lightShaderID     = (lightID & L_TYPE_MASK) + 1;
         uint lightUniformIndex = (lightID & L_INDEX_MASK) >> 4;
   
         Radiance radiance;
   
         {
            matrix LightToWorld = shadePointLight_TransformsArray[lightUniformIndex];
            __shadePointLight_Constants uVars = shadePointLight_ConstantsArray[lightUniformIndex];
            radiance = shadePointLight(uVars.lightIntensityAndRadius, uVars.worldSpacePosition, LightToWorld, surface);
         }
   
         color.xyz += shadeBRDF(surface, radiance);
      }
   }
   return color;
}

[numthreads (16, 8, 1)]
void __UberSurfaceShader
   ( uint3 globalThreadID : SV_DispatchThreadID
   , uint3 tileThreadID   : SV_GroupThreadID
   , uint3 groupID        : SV_GroupID )
{
   uint tileID = __tiles[__offsets[ShaderID] + groupID.x];
   const uint2 PixelCoord = uint2
       ( (tileID & 0x0000ffff) * TILE_SIZE_X + tileThreadID.x
       , (tileID >> 16) * TILE_SIZE_Y + tileThreadID.y);
   const uint PixelIndex = PixelCoord.x + ImageSize.x * PixelCoord.y;
   uint groupThreadID = tileThreadID.x + tileThreadID.y * TILE_SIZE_X;

   if (AA_MODE == SUPER_SAMPLED_AA)
   {
      if (SAMPLES_PER_PIXEL == 1)
      {
         uint vData = __visibilityBuffer.Load(PixelCoord, 0);
         uint tID = vData >> INSTANCE_ID_BITS;
         uint iID = (vData & INSTANCE_ID_MASK);
         uint4 iData  = __instanceToIndexMap[iID];
         uint iOffset = iData.x;
         uint vOffset = iData.y;
         uint sID     = iData.z;
         if (sID != ShaderID) return;

         matrix ObjectToWorld = __transformUberBuffer[iID];
         float2 ImageCoord = float2(PixelCoord.x + 0.5f, PixelCoord.y + 0.5f) / ImageSize;
         float FilterWidth = 1.0f;
         VertexInput vIn = __loadAndInterpolateVertex
            ( iOffset
            , vOffset
            , tID, ObjectToWorld, PixelCoord, FilterWidth);
         float4 c = __shadeDifferentialSurface(vIn, ObjectToWorld, PixelCoord, ImageCoord, FilterWidth);
         __writePixel(PixelIndex, c);
      }
      else
      {
         int sCount = 0;
         float4 color = 0.0f;
         for (int sampleID = 0; sampleID < SAMPLES_PER_PIXEL; ++sampleID)
         {
            uint vData = __visibilityBuffer.Load(PixelCoord, sampleID);
            uint tID = vData >> INSTANCE_ID_BITS;
            uint iID = vData & INSTANCE_ID_MASK;
            uint4 iData  = __instanceToIndexMap[iID];
            uint iOffset = iData.x;
            uint vOffset = iData.y;
            uint sID     = iData.z;
            if (sID != ShaderID) continue;

             sCount++;
            float FilterWidth = 1.0f;
            matrix ObjectToWorld = __transformUberBuffer[iID];
            float2 ImageCoord = float2(PixelCoord.x + 0.5f, PixelCoord.y + 0.5f) / ImageSize;
            VertexInput vIn = __loadAndInterpolateVertex
               ( iOffset
               , vOffset
               , tID, ObjectToWorld, PixelCoord, FilterWidth);
            color += __shadeDifferentialSurface(vIn, ObjectToWorld, PixelCoord, ImageCoord, FilterWidth);
         }
         __blendPixel(PixelIndex, color * INV_SPP);
      }
   }
   else if (AA_MODE == WIREFRAME_AA)
   {
      uint stencil = __msaaStencil[PixelCoord];
      if (stencil == 0) __writePixel(PixelIndex, float4(0.1f, 0.05f, 0.1f, 1.0f));
      else              __writePixel(PixelIndex, float4(1.0f, 1.0f, 1.0f, 1.0f));
   }
   else if (AA_MODE == TRIANGLE_SAMPLED_AA)
   {
      #define PIXEL_COORD x
      #define SUBSAMPLE y
      #define V_SAMPLE z
      #define WEIGHT w
      #define WEIGHTED_COLOR xyz

      uint myShadingSampleIndices[SAMPLES_PER_PIXEL];
      uint numMyShadingSamples = 0;

      uint groupThreadID = tileThreadID.x + tileThreadID.y * TILE_SIZE_X;
      if (groupThreadID == 0) nShadingSamples = 0;
      GroupMemoryBarrierWithGroupSync();

      uint vBuffer[SAMPLES_PER_PIXEL];
      [unroll] for (int a = 0; a < SAMPLES_PER_PIXEL; ++a)
         vBuffer[a] = __visibilityBuffer.Load(PixelCoord, a);

      [unroll] for (int i = 0; i < SAMPLES_PER_PIXEL; ++i)
      {
         uint v_0 = vBuffer[i];
         uint iID_0 = v_0 & INSTANCE_ID_MASK;
         uint tID_0 = v_0 >> INSTANCE_ID_BITS;
         uint sID = __instanceToIndexMap[iID_0].z;
         if (sID == ShaderID)
         {
            float weight = INV_SPP;
            for (int j = i+1; j < SAMPLES_PER_PIXEL; ++j)
            {
               if (vBuffer[j] == v_0)
               {
                  vBuffer[j] = 0xffffffff;
                  weight += INV_SPP;
               }
            }

            int idx;
            InterlockedAdd(nShadingSamples, 1, idx);
            shadingRecords[idx].WEIGHT = asuint(weight);
            shadingRecords[idx].V_SAMPLE = (v_0);
            shadingRecords[idx].PIXEL_COORD = __packPixelCoords(PixelCoord);
            shadingRecords[idx].SUBSAMPLE = i;
            myShadingSampleIndices[numMyShadingSamples++] = idx;
         }
      }

      GroupMemoryBarrierWithGroupSync();

      uint s = groupThreadID;
      while (s < nShadingSamples)
      {
         ShadingRecord r = shadingRecords[s];
         uint vData = r.V_SAMPLE;

         float FilterWidth = 1.0f;
         uint tID = vData >> INSTANCE_ID_BITS;
         uint iID = vData & INSTANCE_ID_MASK;
         uint4 iData  = __instanceToIndexMap[iID];
         uint iOffset = iData.x;
         uint vOffset = iData.y;

         uint2 localPixelCoord = __unpackPixelCoords(shadingRecords[s].PIXEL_COORD);
         matrix ObjectToWorld = __transformUberBuffer[iID];
         float2 ImageCoord = float2(localPixelCoord.x + 0.5f, localPixelCoord.y + 0.5f) / ImageSize;
         VertexInput vIn = __loadAndInterpolateVertex
            ( iOffset
            , vOffset
            , tID, ObjectToWorld, localPixelCoord, FilterWidth);
         float4 c = __shadeDifferentialSurface(vIn, ObjectToWorld, localPixelCoord, ImageCoord, FilterWidth);
         shadingRecords[s].WEIGHTED_COLOR = asuint(asfloat(r.WEIGHT)*c);
         s += TILE_SIZE;
      }

      GroupMemoryBarrierWithGroupSync();
      uint PixelIndex = PixelCoord.x + ImageSize.x * PixelCoord.y;
      float3 pColor = 0.0f;
      for (uint p = 0; p < numMyShadingSamples; ++p)
         pColor += asfloat(shadingRecords[myShadingSampleIndices[p]].WEIGHTED_COLOR);
      if (numMyShadingSamples < SAMPLES_PER_PIXEL)
         __blendPixel(PixelIndex, float4(pColor, 1.0f));
      else
         __writePixel(PixelIndex, float4(pColor, 1.0f));

   }
}
