/******************************************************************************\

                Copyright (c) 2013 Chris Burns & Warren Hunt
                            All rights reserved.

       Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
   
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   * Neither the name of the <organization> nor the
     names of its contributors may be used to endorse or promote products
     derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
    OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
              EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

      This code accompanies the Journal of Computer Graphics Techniques 
  (http://jcgt.org) article titled "The Visibility Buffer: A Cache-Friendly 
    Approach to Deferred Shading". These files were automatically generated 
                    by the system described in the article.

\******************************************************************************/

#define __LANCELOT__

cbuffer __SystemConstants
{
   matrix ObjectToWorld;
   matrix WorldToCamera;
   matrix CameraToClip;    // projection matrix
};

struct VertexInput_Storage
{
   float4 position : POSITION;
   half4 tangent_texCoord_u : TANGENT;
   half4 bitangent_texCoord_v : BINORMAL;
};
struct VertexInput
{
   float4 position  : POSITION;
   float4 tangent   : TANGENT;
   float4 bitangent : BINORMAL;
   float4 texCoord  : TEXCOORD;
};

struct VertexOutput
{
   float4 position  : POSITION;
   float4 tangent   : TANGENT;
   float4 bitangent : BINORMAL;
   float4 texCoord  : TEXCOORD;

   float4 clipSpacePosition : SV_POSITION;
};
VertexInput __unpackVertex(VertexInput_Storage vInStored)
{
   VertexInput vIn;
   vIn.position  = vInStored.position;
   vIn.tangent   = float4(vInStored.tangent_texCoord_u.xyz, 0.0f);
   vIn.bitangent = float4(vInStored.bitangent_texCoord_v.xyz, 0.0f);
   vIn.texCoord  = float4
      ( vInStored.tangent_texCoord_u.w
      , vInStored.bitangent_texCoord_v.w
      , 0.0f, 0.0f);
   return vIn;
}

VertexOutput shadeVertex(VertexInput vIn)
{
   VertexOutput vOut;
   vOut.position  = vIn.position;
   vOut.texCoord  = vIn.texCoord;
   vOut.tangent   = mul( ObjectToWorld, float4(vIn.tangent.xyz, 0.0)  );
   vOut.bitangent = mul( ObjectToWorld, float4(vIn.bitangent.xyz, 0.0));
   return vOut;
}
VertexOutput __VertexShader(VertexInput_Storage vInStored)
{
   VertexInput vIn = __unpackVertex(vInStored);
   VertexOutput vOut = shadeVertex(vIn);
   vOut.clipSpacePosition
      = mul( CameraToClip
           , mul( WorldToCamera
                , mul(ObjectToWorld, vOut.position)));
   return vOut;
}
