/******************************************************************************\

                Copyright (c) 2013 Chris Burns & Warren Hunt
                            All rights reserved.

       Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
   
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   * Neither the name of the <organization> nor the
     names of its contributors may be used to endorse or promote products
     derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
    OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
              EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

      This code accompanies the Journal of Computer Graphics Techniques 
  (http://jcgt.org) article titled "The Visibility Buffer: A Cache-Friendly 
    Approach to Deferred Shading". These files were automatically generated 
                    by the system described in the article.

\******************************************************************************/

#define TILE_SIZE_X          16
#define TILE_SIZE_Y          8
#define PIXELS_PER_TILE      128
#define SAMPLES_PER_PIXEL    4
#define INSTANCE_ID_MASK     1023
#define INSTANCE_ID_BITS     10
#define NUM_CLOSURES         22

cbuffer __SystemConstants
{
   int2 ImageSize;
   int NumTilesAcross;
}

struct WorkItem
{
   uint tile;        /// scanline index of tile
   uint closureID;   /// the shader to execute
};

// shader id is in component z
Buffer<uint4>                    __instanceToIndexMap;
RWBuffer<uint>                   __closureToTileCountMap;
AppendStructuredBuffer<WorkItem> __jobList;
Texture2DMS<uint>                __visibilityBuffer;
RWTexture2D<uint>                __stencil;

groupshared uint scoreboard[NUM_CLOSURES];
[numthreads (TILE_SIZE_X, TILE_SIZE_Y, 1)]
void __GenerateGlobalWorklist
      ( uint3 localID : SV_GroupThreadID
      , uint3 globalID : SV_DispatchThreadID)
{
   const uint2 MyPixel = globalID.xy;
   const uint TileIndex = ((MyPixel.y/TILE_SIZE_Y) << 16) | (MyPixel.x/TILE_SIZE_X);
   const uint LocalThreadID = localID.x + localID.y * TILE_SIZE_X;

   if (LocalThreadID == 0)
   { [unroll] for (int i = 0; i < NUM_CLOSURES; ++i) scoreboard[i] = 0; }
   GroupMemoryBarrierWithGroupSync();

   // isHomogeneous means that the pixel contains exactly one visible triangle
   uint r = __visibilityBuffer.Load(MyPixel, 0);
   [unroll] for (int a = 0; a < SAMPLES_PER_PIXEL; ++a)
   {
      uint v = __visibilityBuffer.Load(MyPixel, a);
      uint iID = v & INSTANCE_ID_MASK;
      scoreboard[__instanceToIndexMap[iID].z] = 1;
   }
   GroupMemoryBarrierWithGroupSync();
   uint c = LocalThreadID;
   while (c < NUM_CLOSURES)
   {
      [branch] if (scoreboard[c] > 0)
      {
         uint x;
         InterlockedAdd(__closureToTileCountMap[c], 1, x);
         WorkItem wItem;
         wItem.tile = TileIndex;
         wItem.closureID = c;
        __jobList.Append(wItem);
      }
      c += PIXELS_PER_TILE;
   }
}

