/******************************************************************************\

                Copyright (c) 2013 Chris Burns & Warren Hunt
                            All rights reserved.

       Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
   
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   * Neither the name of the <organization> nor the
     names of its contributors may be used to endorse or promote products
     derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
    OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
              EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

      This code accompanies the Journal of Computer Graphics Techniques 
  (http://jcgt.org) article titled "The Visibility Buffer: A Cache-Friendly 
    Approach to Deferred Shading". These files were automatically generated 
                    by the system described in the article.

\******************************************************************************/

#define NUM_WORKER_THREADS     256
#define NUM_CLOSURES           22

cbuffer __SystemConstants
{
    uint NumWorkItems;
}

struct WorkItem
{
    uint tile;        /// coordinates of the tile
    uint closureID;   /// the shader to execute
};

RWBuffer<uint>                    __offsets;
RWBuffer<uint>                    __tiles;
Buffer<int>                       __closureToTileCountMap;
ConsumeStructuredBuffer<WorkItem> __jobList;

[numthreads (NUM_WORKER_THREADS, 1, 1)]
void __SortWorklists ( uint3 globalThreadID : SV_DispatchThreadID)
{
   /// initializing the offsets list is a prefix-scan
   /// operation, which can be done in parallel, but for now
   /// we're going to do it serially since initially we do not
   /// expect the length of the list to exceed 64 items

   if (globalThreadID.x == 0)
   {
      __offsets[0] = __offsets[1] = 0;
      for (int i = 2; i < NUM_CLOSURES+1; ++i)
         __offsets[i] = __offsets[i-1] + __closureToTileCountMap[i-2];
   }

   DeviceMemoryBarrierWithGroupSync();

   uint count = globalThreadID.x;
   while (count < NumWorkItems)
   {
      WorkItem item = __jobList.Consume();

      uint idx;
      InterlockedAdd(__offsets[item.closureID+1], 1, idx);
      __tiles[ idx ] = item.tile;

      count += NUM_WORKER_THREADS;
   }
}

